<?php

namespace Teknisa\FIN\Modules\ContaReceber\Services\Retorno\Banco;

use Financeiro\Entities\Cnabrec;
use Geral\Entities\Cliente;
use Geral\Entities\Empresa;
use Financeiro\Entities\Boletreceber;
use Teknisa\FIN\Modules\ContaReceber\Services\Retorno\LeitorRetorno;
use Teknisa\FIN\Modules\ContaReceber\Services\Retorno\Retorno;

class LeitorRetornoBradesco extends LeitorRetorno{

    const CODIGO_COMPENSACAO_BRADESCO = "237";
    CONST ID_TP_ARQUIVO_HEADER_ARQUIVO = '0';
    CONST ID_TP_ARQUIVO_HEADER_LOTE = '1';
    CONST ID_TP_TRANSACAO_1 = '1';
    CONST ID_TP_TRANSACAO_2 = '2';
    CONST ID_TP_ARQUIVO_TRAILER_LOTE = '5';
    CONST ID_TP_TRAILER_ARQUIVO = '9';
    CONST TIPO_INS_JUR_CPF = 1;
    CONST TIPO_INS_JUR_CNPJ = 2;
    CONST RETORNO = 2;
    CONST SEGMENTO_A = 'A';
    CONST SEGMENTO_J = 'J';
    CONST SEGMENTO_O = 'O';
    CONST SEGMENTO_N = 'N';

    private $arquivo;

    private $cdagencia;
    private $cdconta;
    private $ocorrenciaHeader;
    private $contHeaderLote = 0;
    /** @var $retorno Retorno */
    private $retorno;

    /**
     * @param $arquivo
     * @param $cdempresa
     * @param $cdbanco
     * @throws \Exception
     */
    public function __construct($arquivo, $cdempresa, $cdbanco, $nrorg){
        try {
            $this->cdbanco = $cdbanco;
            $this->arquivo = $arquivo;
            $this->cdempresa = $cdempresa;
            $this->nrorg = $nrorg;
        } catch (\Exception $e) {
            throw $e;
        }
    }

    public function getEstruturaArquivo(){
        return array(
            self::ID_TP_ARQUIVO_HEADER_ARQUIVO,
            self::ID_TP_ARQUIVO_HEADER_LOTE,
            self::ID_TP_TRANSACAO_1,
            self::ID_TP_ARQUIVO_TRAILER_LOTE,
            self::ID_TP_TRAILER_ARQUIVO
        );
    }

    public function leArquivo(){

        try {
            foreach ($this->arquivo as $linha) {
                $cont = $this->contlinha++;
                if ($linha){
                    $idsegmento = substr($linha, 13, 1);
                    if($idsegmento == 'U')
                        $cont = $cont - 1;
                    else
                        $this->retorno = new Retorno();

                    // Tamanho do Arquivo - 240 ou 400
                    $tamanhoLinha = strlen($linha);

                    // Busca o Tipo de Registro do Arquivo
                    $idRegistro =  $idRegistro = $tamanhoLinha == 400?substr($linha, 0, 1):substr($linha, 7, 1);
                    
                    if(self::validaArquivo($linha, self::ID_TP_ARQUIVO_HEADER_ARQUIVO, self::getEstruturaArquivo(), $this->contlinha)) {
                        switch ($idRegistro) {

                            case(self::ID_TP_ARQUIVO_HEADER_ARQUIVO):
                                self::leHeaderArquivo($linha);
                                break;

                            case(self::ID_TP_TRANSACAO_1):
                                
                                self::leTransacao($linha);
                                break;
                        }
                    }
                    $this->retornos[$cont] = $this->retorno;
                }
            }
            self::preparaOcorrencias($this->cdbanco, $this->nrorg);

        } catch (\Exception $e) {
            throw $e;
        }
    }


    /**
     * @param $cdempresa
     * @throws \Exception
     * @return
     */
    private function getIncricaoEmpresa($cdempresa){
        /** @var $empresa Empresa */
        $empresa = Empresa::findOneByNrorgAndCdempresa($this->nrorg, $cdempresa);
        if($empresa)
            return $empresa->getNrinsjurempr();
        else
            throw new \Exception ('Empresa não Encontrada.');
    }


    /**
     * // =====================================================================  //
    // ====================  REGISTRO HEADER  DE ARQUIVO  ====================  //
    // =====================================================================  //
     * @param TXT $header
     */
    private function leHeaderArquivo($header){

        // CAMPO: Código do BCO na compensação
        // POSIÇÃO: 1 a 3
        // PICTURE: 9(3)
        if(substr($header, 76, 3) != self::CODIGO_COMPENSACAO_BRADESCO) {
            self::gravaErros($this->contlinha, 'Número do banco no arquivo é diferente do número do banco informardo no retorno.', true);
        }

        // CAMPO: Identificação do Arquivo de Retorno
        // POSIÇÃO: 143 a 143
        // PICTURE: 9(1)
        if (substr($header, 1, 1) != '2'){
            self::gravaErros($this->contlinha, 'Arquivo informado não é um arquivo de retorno', true);
        }
    }

    /**
    // =====================================================================  //
    // ====================      REGISTRO DETALHE       ====================  //
    // =====================================================================  //
     * @param $detalhe
     */
    private function leTransacao($detalhe){

        // CAMPO: Identificação do Título no Banco
        // POSIÇÃO: 071 a 082
        // PICTURE: X(12)
        $cdnosnumrec = trim(substr($detalhe, 70, 12));

        // CAMPO: Nº controle do participante - chave do boleto
        // POSIÇÃO: 038 a 062
        // PICTURE: X(25)
        $cdfilial = trim(substr($detalhe, 37, 4));
        $nrlanbolrec = trim(substr($detalhe, 41, 10));

        // CAMPO: Tipo de Inscrição da empresa
        // POSIÇÃO: 002 a 003
        // PICTURE: 9(2)
        $idtpinscempr = trim(substr($detalhe, 1, 2));

        // CAMPO: Inscrição da empresa
        // POSIÇÃO: 004 a 017
        // PICTURE: 9(14)
        $inscjurempr = trim(substr($detalhe, 3, 14));
        $cdagencia = trim(substr($detalhe, 24, 5));
        $cdbanco = '237';
        $cdconta = trim(substr($detalhe, 29, 7));

        // CAMPO: Identificação da Ocorrência no Banco
        // POSIÇÃO: 109 a 110
        // PICTURE: 9(2)
        $idocorrencia = trim(substr($detalhe, 108, 2));
        $cdempresa = $this->cdempresa;
        /** @var $entitieBoleto \Financeiro\Repositories\Boletreceber */
        $entitieBoleto = Boletreceber::getRepository();
        $nossonumero = $entitieBoleto->getNossoNumero($cdnosnumrec, $cdempresa);
        /** @var $boleto Boletreceber */
        $boleto = Boletreceber::findOneByNrlanbolrecAndCdfilial($nrlanbolrec, $cdfilial);
        if(empty($idtpinscempr))
            $this->gravaErros($this->contlinha, 'Tipo de inscrição jurídica inválido.', true);
        if(empty($inscjurempr))
            $this->gravaErros($this->contlinha, 'Inscrição jurídica inválida', true);
        if($nossonumero)
            $nossonumero = trim($nossonumero[0]['CDNOSNUMREC']);
        if(!empty($cdfilial) && !empty($nrlanbolrec) || !empty($nossonumero)){
            if(empty($nossonumero))
                $this->gravaErros($this->contlinha, 'Nosso número inválido.', true);
            if(empty($idocorrencia))
                $this->gravaErros($this->contlinha, 'Identificação da ocorrência inválido.', true);
            $this->retorno->setBoleto('BOLETO');
            if($boleto){
                $this->retorno->setCdfilial($cdfilial);
                $this->retorno->setNrlanbolrec($nrlanbolrec);
                $vencimento = $boleto->getDtvenbolrec();
                $this->retorno->setDtvencimento($vencimento->format('d/m/Y'));
                $this->retorno->setCdreducont($boleto->getCdreducont());
                $this->retorno->setCdtpbaixarec($boleto->getCdtpbaixarec());
                $this->retorno->setVrfinalbrec($boleto->getVrfinalbrec());
                $this->retorno->setIdstatrec($boleto->getIdstatbrec());
            }
            else{
                /** @var $boleto Boletreceber */
                $boleto = Boletreceber::findOneByCdnosnumrec($nossonumero);
                if($boleto){
                    $this->retorno->setCdfilial($boleto->getCdfilial());
                    $this->retorno->setNrlanbolrec($boleto->getNrlanbolrec());
                    $vencimento = $boleto->getDtvenbolrec();
                    $this->retorno->setDtvencimento($vencimento->format('d/m/Y'));
                    $this->retorno->setCdreducont($boleto->getCdreducont());
                    $this->retorno->setCdtpbaixarec($boleto->getCdtpbaixarec());
                    $this->retorno->setVrfinalbrec($boleto->getVrfinalbrec());
                    $this->retorno->setIdstatrec($boleto->getIdstatbrec());
                }
            }
            if(!empty($nossonumero))
                $this->retorno->setNossonumero($nossonumero);
            /** @var $empresa Empresa */
            $empresa = Empresa::findOneByNrinsjurempr($inscjurempr);
            if($empresa)
                $this->retorno->setNmrazsocclie($empresa->getNmrazsocempr());
            $this->retorno->setCdagencia($cdagencia);
            $this->retorno->setCdconta($cdconta);
            
            $vencimento = substr($detalhe, 146, 2) . substr($detalhe, 148, 2) . substr(date("Y"), 0, 2) . substr($detalhe, 150, 2);
            
            $this->retorno->setDtvencimento($vencimento);

            // CAMPO: Valor do Título
            // POSIÇÃO: 153 a 165
            // PICTURE: 9(13)
            $this->retorno->setVrnominal(round(floatval(substr($detalhe, 152, 13)) / 100, 2));
            
            if($boleto instanceof Boletreceber) {
                /** @var $cnab Cnabrec */
                $cnab = Cnabrec::findOneByCdempresaAndNrcnabrec($cdempresa, $boleto->getNrcnabrecrem());
                if($cnab){
                    if($cnab->getIdtipcnabrec() == '0'){
                        $this->retorno->setVrdesconto(
    
                        // CAMPO: Despesas de Cobrança para Ocorrência(02/28)
                        // POSIÇÃO: 176 a 188
                        // PICTURE: 9(13)
                            round(floatval(substr($detalhe, 175, 13)) / 100, 2) +
    
                            // CAMPO: Outras despesas
                            // POSIÇÃO: 189 a 201
                            // PICTURE: 9(13)
                            round(floatval(substr($detalhe, 188, 13)) / 100, 2) +
    
                            // CAMPO: IOF Devido
                            // POSIÇÃO: 215 a 227
                            // PICTURE: 9(13)
                            round(floatval(substr($detalhe, 214, 13)) / 100, 2) +
    
                            // CAMPO: Abatimento concedito sobre o título
                            // POSIÇÃO: 228 a 240
                            // PICTURE: 9(13)
                            round(floatval(substr($detalhe, 227, 13)) / 100, 2) +
    
                            // CAMPO: Desconto concedido
                            // POSIÇÃO: 241 a 253
                            // PICTURE: 9(13)
                            round(floatval(substr($detalhe, 240, 13)) / 100, 2)
                        );
                        $this->retorno->setVracrescimo(
                        // CAMPO: Juros operação em atraso
                        // POSIÇÃO: 202 a 214
                        // PICTURE: 9(13)
                            round(floatval(substr($detalhe, 201, 13)) / 100, 2) +
    
                            // CAMPO: Juros de Mora
                            // POSIÇÃO: 267 a 279
                            // PICTURE: 9(13)
                            round(floatval(substr($detalhe, 266, 13)) / 100, 2) +
    
                            // CAMPO: Outros Créditos
                            // POSIÇÃO: 280 a 292
                            // PICTURE: 9(13)
                            round(floatval(substr($detalhe, 279, 13)) / 100, 2)
                        );
                    }
                }
            }

            // CAMPO: Valor pago
            // POSIÇÃO: 254 a 266
            // PICTURE: 9(11)V9(2)
            $this->retorno->setVrpago(round(floatval(substr($detalhe, 253, 13)) / 100, 2));
            $this->retorno->setCdlinha(str_pad($this->contlinha, 4, '0', STR_PAD_LEFT));
            $cdocorrencia = self::situacaoCnab($idocorrencia);
            $this->retorno->setIdsitucnabr($cdocorrencia);
            
            if($cdocorrencia == '90') {
                // CAMPO: Data do Crédito
                // POSIÇÃO: 296 a 301
                // PICTURE: 9(6)
                $dtpagamento = substr($detalhe, 295, 6);
                $data = substr($dtpagamento, 0, 2) . substr($dtpagamento, 2, 2) . substr(date("Y"), 0, 2) . substr($dtpagamento, 4, 2);
                $this->retorno->setDtpagamento($data);
            }

            // CAMPO: Erros - Registros rejeitados ou alegação do sacado
            // POSIÇÃO: 319 a 328
            // PICTURE: X(8)
            $idErros = substr($detalhe, 318, 10);
            //GRAVA OCORRÊNCIAS DO CNAB
            if($cdocorrencia == '01'){
                if($idocorrencia == '24' or $idocorrencia == '25'){
                    if(substr($idErros, 0, 2) != '')
                        $this->retorno->setOcorrencia($idocorrencia . substr($idErros, 0, 4));
                    if(substr($idErros, 2, 2) != '')
                        $this->retorno->setOcorrencia($this->retorno->getOcorrencia() . substr($idErros, 4, 4));
                }
                else{
                    if(substr($idErros, 0, 2) != '' && substr($idErros, 0, 2) != '00') {
                        $this->retorno->setOcorrencia($idocorrencia . substr($idErros, 0, 2));
                    }
                    if(substr($idErros, 2, 2) != '' && substr($idErros, 2, 2) != '00') {
                        $this->retorno->setOcorrencia($this->retorno->getOcorrencia() . $idocorrencia . substr($idErros, 2, 2));
                    }
                    if(substr($idErros, 4, 2) != '' && substr($idErros, 4, 2) != '00') {
                        $this->retorno->setOcorrencia($this->retorno->getOcorrencia() . $idocorrencia . substr($idErros, 4, 2));
                    }
                    if(substr($idErros, 6, 2) != '' && substr($idErros, 6, 2) != '00') {
                        $this->retorno->setOcorrencia($this->retorno->getOcorrencia() . $idocorrencia . substr($idErros, 6, 2));
                    }
                    if(substr($idErros, 8, 2) != '' && substr($idErros, 8, 2) != '00') {
                        $this->retorno->setOcorrencia($this->retorno->getOcorrencia() . $idocorrencia . substr($idErros, 8, 2));
                    }
                }
            }
        }
    }


    /**
     * @param $cdocorrencia
     * @return string
     */
    private function situacaoCnab($cdocorrencia){
        if($cdocorrencia == '09' || $cdocorrencia == '10')
            return '00'; //Não Agendado
        else if($cdocorrencia == '02' or $cdocorrencia == '14' or $cdocorrencia == '19' or $cdocorrencia == '23')
            return '20'; //Retorno OK
        else if($cdocorrencia == '03' or $cdocorrencia == '24' or $cdocorrencia == '27' or $cdocorrencia == '32')
            return '01'; //Retorno com Erro
        else if($cdocorrencia == '06' or $cdocorrencia == '15' or $cdocorrencia == '17')
            return '90'; //Baixado
        return '10'; //Sem retorno
    }
}
